/**************************************************************************   
 *                                                                        *
 *    File Name : inv.c                                                   *
 *                                                                        *
 *    Author    : KIM, DONGMIN (kimd@top.cis.syr.edu)                     *
 *                                                                        *
 *    Function  : This program is a master program for computing inverse  *
 *                matrix for VDCE application.                            *
 *                This program read data from buffer which stores         *
 *                informations received by vdce_recv.                     *
 *                After the computation, it stores inverse matrix to      *
 *                buffer so that they can be sent by vdce_send.           *
 *                                                                        *
 **************************************************************************/

#include <stdio.h>
#include <malloc.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>

#include "vdce.h"

#define inter_comm "??"

extern	char	*vdce_get_my_name();
extern	char	*vdce_input_name();
extern	char	*vdce_output_name();
extern	int	vdce_data_represent();

vdce_run(argc,argv)
     int  argc;
     char *argv[];
{
  int nsize;
  int inch1, outch1;
  char	**buf;
  char tmp[10];
  char *temp;
  int i, j, k;
  int matrix_size, type, size;
  int nslave, my_id, to, row, load, from, column;
  int LU;
  float **Matrix, **I;
  float L, U, divide, mult;
  float element, zero;
  char	*out1_msg;

  long  total_elapsed;
  FILE  *fd, *fd1;
  char  *task_name;
  int   task_id;
  char  file_name[100];
  char  *io_file;
  int   *counter;
  char  *buf2;
  int   *tmp_matrix_size;
  int   total_size;
  char  *send_buf;
  float *tmp_float;

  timer_reset();

  task_name = (char *) malloc(100*sizeof(char ));
  task_name = vdce_get_my_name();
  task_id = vdce_get_my_id();

  sprintf(file_name, "/tmp/%d_%s", task_id, task_name);

  fd = fopen(file_name, "w");

  inch1 = atoi(argv[1]);
  outch1 = atoi(argv[2]);

  fprintf(fd, "in_port1 = %d, out_port1 = %d\n", inch1,outch1);


  io_file = (char *) malloc(100*sizeof(char ));
  bzero(io_file, 100);

  io_file = vdce_input_name(task_id, inch1);
  if (io_file == NULL) {
    fprintf(fd, "malloc error at io_file");
    fflush(fd);
  }
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */
    fprintf(fd,"vdce receive\n"); 
    fflush(fd);

    nsize=vdce_recv(inch1, &buf2);

    if ( vdce_data_represent(inch1, INPUT) ) { /* different data type */

      temp = strtok((char * ) buf2, " ");
      matrix_size = atoi(temp);
      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }
      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
	if (Matrix[i] == NULL) {
	  fprintf(fd, "malloc error at Matrix[%d]", i);
	}
      }
      for ( i = 1 ; i <= matrix_size ; i++ ) {
	temp = strtok( 0, " ");
	Matrix[i-1][0] = (float)atof(temp);
	for ( j = 1 ; j < matrix_size ; j++ ) {
	  temp = strtok( 0, " ");
	  Matrix[i-1][j] = (float)atof(temp);
	}
      }
    }
    else { /* same data type */

      tmp_matrix_size = (int *)buf2;

      matrix_size = *tmp_matrix_size;

      fprintf(fd, "receivec size : %d   Matrix size : %d\n", nsize, matrix_size);
      fflush(fd);

      Matrix =  (float **) malloc(matrix_size * sizeof(float *));
      if (Matrix == NULL) {
	fprintf(fd, "malloc error at Matrix");
      }

      for ( i = 0 ; i < matrix_size ; i++ ) {
	Matrix[i] = (float *)(buf2 + sizeof(int) + i * sizeof(float) * matrix_size);
      }

      if (nsize <= 0 ) {
	fprintf(fd,"data has not been arrived\n"); 
	fflush(fd);
	exit(1);
      }
    }
  }
  else {
    /* external file */
    /* get input data from file */
    fprintf(fd,"external file\n"); 
    fflush(fd);
    fd1 = fopen(io_file, "r");
 
    if ( fd1 == NULL ) {
      printf("\ncannot find file!\n");
      exit(1);
    }
 
    /* read the size of the matrix. */
    fscanf( fd1, "%d", &matrix_size );

    Matrix =  (float **) malloc(matrix_size * sizeof(float *));
    if (Matrix == NULL) {
      fprintf(fd, "malloc error at Matrix");
    }
    for ( i = 0 ; i < matrix_size ; i++ ) {
      Matrix[i] = (float *) malloc(matrix_size * sizeof(float));
      if (Matrix[i] == NULL) {
        fprintf(fd, "malloc error at Matrix[%d]", i);
      }
    }

    /* read from the file */
    for ( i = 0 ; i < matrix_size ; i++ ) {
      for ( j = 0 ; j < matrix_size ; j++ ) {
	fscanf(fd1, "%f", &element);
	Matrix[i][j] = element;
      }
    }
    fclose(fd1);
  }

  I =  (float **) malloc(matrix_size * sizeof(float *));
  for ( i = 0 ; i < matrix_size ; i++ ) {
    I[i] = (float *) malloc(matrix_size * sizeof(float));
  }

  /* Initiate the Identity matrix */
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( i == j ) {
	I[i][j] = 1.0;
      }
      else {
	I[i][j] = 0.0;
      }
    }
  }

  L = 0.;
  U = 0.;

  /* check if the matrix is L, U or neither */
  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      if ( i < j ) {
	L = L + Matrix[i][j];
      }
      else if ( i > j ) {
	U = U + Matrix[i][j];
      }
      else {
      }
    }
  }

  if ( L == 0.0 ) { /* the matrix is lower matrix */

    LU = 1;
  }
  else if ( U == 0.0 ) {  /* the matrix is upper matrix */
    LU = 2;
  }
  else { /* neither */
    LU = 0;
  }

    /* convert the matrix to upper matrix if it is neither L nor U */
  if ( LU == 0 ) {
    LU == 2;
    for ( i = 0 ; i < matrix_size - 1 ; i++ ) {
      for ( j = i + 1 ; j < matrix_size ; j++ ) {
	divide = Matrix[j][i];
	mult = Matrix[i][i];
	if ( divide != 0 ) {
	  for ( k = 0 ; k < matrix_size ; k++ ) {

	    Matrix[j][k] = Matrix[i][k] - ((mult / divide) * Matrix[j][k]);

	    I[k][j] = I[k][i] - ((mult / divide) * I[k][j]);
	  }
	}
	Matrix[j][i] = 0.;
      }
    }
  }

  fprintf(fd, "Matrix\n");
  fflush(fd);

  compute_inverse(Matrix, I, matrix_size, LU);

  fprintf(fd, "\nResult Matrix\n");
  fflush(fd);

  bzero(io_file, 100);
  io_file = vdce_output_name(task_id, outch1);
  fprintf(fd,"io file name : %s\n", io_file); 
  fflush(fd);

  if ( !strcmp(io_file, inter_comm) ) {
    /* inter process communication */
    /* VDCE receive */

    if ( vdce_data_represent(outch1, OUTPUT) ) { /* different data type */

      /* write to the buffer */
      send_buf = (char *) malloc((matrix_size * matrix_size * 10 + 10) * sizeof(char));
      if (send_buf == NULL) {
        fprintf(fd, "malloc error at send_buf");
      }
      total_size = sizeof(send_buf);
      bzero(tmp, 10);
      bzero(send_buf, (matrix_size * matrix_size * 10 + 10));
      sprintf(tmp, "%d ", matrix_size);
      strcat(send_buf, tmp);
      element = 1.0;
      zero = 0.0;
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  bzero(tmp, 10);
	  sprintf( tmp, "%5.3f ", Matrix[i][j]);
	  strcat(send_buf, tmp);
	}
      }
      strcat(send_buf,"\n");
    }
    else {

      total_size = sizeof(matrix_size) + sizeof(float) * matrix_size * matrix_size;

      fprintf(fd, "total size : %d\n", total_size);
      fflush(fd);

      send_buf = (char *)malloc(total_size + 1);

      bcopy(&matrix_size, send_buf, sizeof(matrix_size));

      for ( i = 0 ; i < matrix_size ; i++ ) {
	bcopy(Matrix[i], send_buf + sizeof(matrix_size) + i * sizeof(float) * matrix_size, sizeof(float) * matrix_size);
      }

    }
    vdce_send(outch1, send_buf, total_size);
  }
  else {
    /* external file */
    fd1 = fopen(io_file, "w");
 
    if ( fd1 == NULL ) {
      fprintf(fd, "\ncannot find file!\n");
    }
    else {
      fprintf(fd1, "%d\n", matrix_size);
      for ( i = 0 ; i < matrix_size ; i++ ) {
	for ( j = 0 ; j < matrix_size ; j++ ) {
	  fprintf( fd1, "%7.3f", Matrix[i][j]);
	}
	fprintf(fd1, "\n");
      }
      fclose(fd1);
    }
  }

  total_elapsed = timer_elapsed();

  fprintf(fd,"%s%d  %d\n", task_name, task_id, total_elapsed);
  fclose(fd);

  report_execution_time(2,total_elapsed);
}

vdce_def(mytaskinfo)
     VDCE_TASKINFO	*mytaskinfo;
{

  vdce_set_taskname(mytaskinfo,"INV");
	
  vdce_set_inputs(mytaskinfo,"L");

  vdce_set_outputs(mytaskinfo,"INV_L");
}

compute_inverse(Matrix, I, matrix_size, LU)
float **Matrix, **I;
int  matrix_size, LU;
{
  int	 i, j, k, step, type, from, to;

  int	 current_id, current_row, nslave;
  int	 total_size;
  float **inverse;
  float divisor, sum;
  int	 current, current_slave;


  inverse = (float **) malloc(matrix_size * sizeof(float *));
  for ( i = 0 ; i < matrix_size ; i++ )
    inverse[i] = (float *) malloc(matrix_size * sizeof(float));

  /* Warning :
     the inverse matrix of L (or U) is also L (or U).
     So we can reduce the computation time if we notice this.
  */
  if ( LU == 1 ) { /* L matrix */

    for ( k = 0 ; k < matrix_size ; k++ ) {
      for ( i = 0 ; i < matrix_size ; i++ ) {
	sum = 0.;
	for ( j = 0 ; j < i ; j++ ) {
	  sum = sum + Matrix[i][j] * inverse[k][j];
	}
	inverse[k][i] = I[i][k] - sum;
      }
    }
  }
  else {

    for ( k = 0 ; k < matrix_size ; k++ ) {
      for ( i = matrix_size - 1 ; i >= 0 ; i-- ) {
	divisor = Matrix[i][i];
	sum = 0.;
	for ( j = i + 1 ; j < matrix_size ; j++ ) {
	  sum = sum + Matrix[i][j] * inverse[k][j];
	}
	inverse[k][i] = ( I[i][k] - sum ) / divisor;
      }
    }
  }

  for ( i = 0 ; i < matrix_size ; i++ ) {
    for ( j = 0 ; j < matrix_size ; j++ ) {
      Matrix[i][j] = inverse[i][j];
    }
  }

}

